#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""Filter functions for jinja"""

import rich


def jinja_filters():
    """return a dictionary of all the filters

    suitable for use by:

    env = jinja2.Environment()
    env.filters = all_filters()
    """
    filters = {}
    filters["fg_hex"] = fg_hex
    filters["fg_hex_no_hash"] = fg_hex_no_hash
    filters["fg_rgb"] = fg_rgb
    filters["bg_hex"] = bg_hex
    filters["bg_hex_no_hash"] = bg_hex_no_hash
    filters["bg_rgb"] = bg_rgb
    filters["ansi_on"] = ansi_on
    filters["ansi_off"] = ansi_off
    return filters


def fg_hex(value):
    """when applied to a style, get the hex value of the foreground color"""
    if isinstance(value, rich.style.Style):
        if value.color and not value.color.is_default:
            return value.color.get_truecolor().hex
        # empty or default styles don't have a color
        return ""
    # filter applied to something that wasn't a style, no-op
    return value


def fg_hex_no_hash(value):
    """when applied to a style, get the hex value (minus the hash mark)
    of the foreground color"""
    if isinstance(value, rich.style.Style):
        if value.color and not value.color.is_default:
            return value.color.get_truecolor().hex.replace("#", "")
        # empty or default styles don't have a color
        return ""
    # filter applied to something that wasn't a style, no-op
    return value


def fg_rgb(value):
    """when applied to a style, get a rgb function representation
    of the foreground color"""
    if isinstance(value, rich.style.Style):
        if value.color and not value.color.is_default:
            return value.color.get_truecolor().rgb
        # empty or default styles don't have a color
        return ""
    # filter applied to something that wasn't a style, no-op
    return value


def bg_hex(value):
    """when applied to a style, get the hex value of the background color"""
    if isinstance(value, rich.style.Style):
        if value.bgcolor and not value.bgcolor.is_default:
            return value.bgcolor.get_truecolor().hex
        # not all styles have a background color, if they don't
        # or if it's the default color, return an empty string
        return ""
    # filter applied to something that wasn't a style, no-op
    return value


def bg_hex_no_hash(value):
    """when applied to a style, get the hex value (minus the hash mark)
    of the foreground color"""
    if isinstance(value, rich.style.Style):
        if value.bgcolor and not value.bgcolor.is_default:
            return value.bgcolor.get_truecolor().hex.replace("#", "")
        # not all styles have a background color, if they don't
        # or if it's the default color, return an empty string
        return ""
    # filter applied to something that wasn't a style, no-op
    return value


def bg_rgb(value):
    """when applied to a style, get a rgb function representation
    of the background color"""
    if isinstance(value, rich.style.Style):
        if value.bgcolor and not value.bgcolor.is_default:
            return value.bgcolor.get_truecolor().rgb
        # not all styles have a background color, if they don't
        # or if it's the default color, return an empty string
        return ""
    # filter applied to something that wasn't a style, no-op
    return value


def ansi_on(value):
    """when applied to a style, get the ansi codes to tell the terminal
    to start printing text in that style"""
    if isinstance(value, rich.style.Style):
        splitter = "-----"
        ansistring = value.render(splitter)
        out, _ = ansistring.split(splitter)
        return out
    return value


def ansi_off(value):
    """when applied to a style, get the ansi codes to tell the terminal
    to stop printing text in that style"""
    if isinstance(value, rich.style.Style):
        splitter = "-----"
        ansistring = value.render(splitter)
        _, out = ansistring.split(splitter)
        return out
    return value



#
# Copyright (c) 2025 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

import pytest
import rich

import dye
from dye import Dye, Pattern

#
# test the ls_colors agent
#

# we only reallly have to test that the style name maps to the right code in ls_colors
# ie directory -> di, or setuid -> su. The ansi codes are created by rich.style
# so we don't really need to test much of that
STYLE_TO_LSCOLORS = [
    ("text", "", ""),
    ("text", "default", "no=0"),
    ("file", "default", "fi=0"),
    ("directory", "#8be9fd", "di=38;2;139;233;253"),
    ("symlink", "green4 bold", "ln=1;38;5;28"),
    ("multi_hard_link", "blue on white", "mh=34;47"),
    ("pipe", "#f8f8f2 on #44475a underline", "pi=4;38;2;248;248;242;48;2;68;71;90"),
    ("so", "bright_white", "so=97"),
    ("door", "bright_white", "do=97"),
    ("block_device", "default", "bd=0"),
    ("character_device", "black", "cd=30"),
    ("broken_symlink", "bright_blue", "or=94"),
    ("missing_symlink_target", "bright_blue", "mi=94"),
    ("setuid", "bright_blue", "su=94"),
    ("setgid", "bright_red", "sg=91"),
    ("sticky", "blue_violet", "st=38;5;57"),
    ("other_writable", "blue_violet italic", "ow=3;38;5;57"),
    ("sticky_other_writable", "deep_pink2 on #ffffaf", "tw=38;5;197;48;2;255;255;175"),
    ("executable_file", "cornflower_blue on grey82", "ex=38;5;69;48;5;252"),
    ("file_with_capability", "red on black", "ca=31;40"),
]


@pytest.mark.parametrize("name, styledef, expected", STYLE_TO_LSCOLORS)
def test_ls_colors_from_style(name, styledef, expected):
    style = rich.style.Style.parse(styledef)
    # we have to have a pattern in order for the agent to initialize
    # so lets make a fake one
    pattern_str = """
    [scopes.myscope]
    agent = "ls_colors"
    """
    pattern = Pattern.loads(pattern_str)
    agent = dye.agents.LsColors(pattern.scopes["myscope"])
    # now we can go test the render
    code, render = agent.ls_colors_from_style(
        name,
        style,
        agent.LS_COLORS_MAP,
        "myscope",
        allow_unknown=False,
    )
    assert render == expected
    assert code == expected[0:2]


def test_ls_colors_no_styles(dye_cmdline, capsys):
    pattern_str = """
        [scopes.lsc]
        agent = "ls_colors"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert out == 'export LS_COLORS=""\n'


def test_ls_colors_unknown_style(dye_cmdline, capsys):
    pattern_str = """
        [scopes.lsc]
        agent = "ls_colors"
        styles.bundleid = "default"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_ERROR
    assert "unknown style" in err
    assert "lsc" in err


def test_ls_colors_environment_variable(dye_cmdline, capsys):
    pattern_str = """
        [scopes.lsc]
        agent = "ls_colors"
        environment_variable = "OTHER_LS_COLOR"
        styles.file = "default"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert out == 'export OTHER_LS_COLOR="fi=0"\n'


def test_ls_colors_styles_variables(dye_cmdline, capsys):
    pattern_str = """
        [variables]
        pinkvar = "magenta3"

        [styles]
        warning = "yellow on red"

        [scopes.lsc]
        agent = "ls_colors"
        styles.file = "{{styles.warning}}"
        styles.directory = "{{variables.pinkvar}}"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert out == 'export LS_COLORS="fi=33;41:di=38;5;164"\n'


def test_ls_colors_clear_builtin(dye_cmdline, capsys):
    pattern_str = """
        [scopes.lsc]
        agent = "ls_colors"
        clear_builtin = true
        styles.directory = "bright_blue"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    expected = (
        'export LS_COLORS="di=94:no=0:fi=0:ln=0:'
        "mh=0:pi=0:so=0:do=0:bd=0:cd=0:or=0:mi=0:"
        'su=0:sg=0:st=0:ow=0:tw=0:ex=0:ca=0"\n'
    )
    assert out == expected


def test_ls_colors_clear_builtin_not_boolean(dye_cmdline, capsys):
    pattern_str = """
        [scopes.lsc]
        agent = "ls_colors"
        clear_builtin = "error"
        style.directory = "bright_blue"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_ERROR
    assert not out
    assert "'clear_builtin' to be true or false" in err



#
# Copyright (c) 2025 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

import pytest

from dye import Dye


#
# test the iterm agent
#
def test_profile(dye_cmdline, capsys):
    pattern_str = """
    [scopes.iterm]
    agent = "iterm"
    cursor = "box"
    profile = "superduper"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    lines = out.splitlines()
    # we have multiple directives in this scope, but the profile directive
    # should always come out first
    assert len(lines) == 2
    # rf'...' lets us use f string interpolation, but the r disables
    # escape processing, just what we need for this test
    assert lines[0] == r'builtin echo -en "\e]1337;SetProfile=superduper\a"'


def test_tab(dye_cmdline, capsys):
    pattern_str = """
    [scopes.iterm]
    agent = "iterm"
    styles.tab = "#337799"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    lines = out.splitlines()
    assert len(lines) == 3
    assert lines[0] == r'builtin echo -en "\e]6;1;bg;red;brightness;51\a"'
    assert lines[1] == r'builtin echo -en "\e]6;1;bg;green;brightness;119\a"'
    assert lines[2] == r'builtin echo -en "\e]6;1;bg;blue;brightness;153\a"'


def test_tab_default(dye_cmdline, capsys):
    pattern_str = """
    [scopes.iterm]
    agent = "iterm"
    styles.tab = "default"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    lines = out.splitlines()
    assert len(lines) == 1
    assert lines[0] == r'builtin echo -en "\e]6;1;bg;*;default\a"'


def test_foreground(dye_cmdline, capsys):
    pattern_str = """
    [scopes.iterm]
    agent = "iterm"
    styles.foreground = "#ffeebb"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, _ = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    lines = out.splitlines()
    assert len(lines) == 1
    assert lines[0] == r'builtin echo -en "\e]1337;SetColors=fg=ffeebb\a"'


def test_background(dye_cmdline, capsys):
    pattern_str = """
    [scopes.iterm]
    agent = "iterm"
    styles.background = "#221122"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, _ = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    lines = out.splitlines()
    assert len(lines) == 1
    assert lines[0] == r'builtin echo -en "\e]1337;SetColors=bg=221122\a"'


def test_cursor(dye_cmdline, capsys):
    pattern_str = """
    [scopes.iterm]
    agent = "iterm"
    cursor = "underline"
    styles.cursor = "#cab2cd"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    lines = out.splitlines()
    assert len(lines) == 2
    assert lines[0] == r'builtin echo -en "\e]1337;CursorShape=2\a"'
    assert lines[1] == r'builtin echo -en "\e]1337;SetColors=curbg=cab2cd\a"'


CURSORS = [
    ("block", "0"),
    ("box", "0"),
    ("vertical_bar", "1"),
    ("vertical", "1"),
    ("bar", "1"),
    ("pipe", "1"),
    ("underline", "2"),
]


@pytest.mark.parametrize("name, code", CURSORS)
def test_cursor_shape(dye_cmdline, capsys, name, code):
    pattern_str = f"""
    [scopes.iterm]
    agent = "iterm"
    cursor = "{name}"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    lines = out.splitlines()
    assert len(lines) == 1
    # rf'...' lets us use f string interpolation, but the r disables
    # escape processing, just what we need for this test
    assert lines[0] == rf'builtin echo -en "\e]1337;CursorShape={code}\a"'


def test_cursor_shape_invalid(dye_cmdline, capsys):
    pattern_str = """
    [scopes.iterm]
    agent = "iterm"
    cursor = "ibeam"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_ERROR
    assert not out
    assert "unknown cursor" in err


def test_cursor_profile(dye_cmdline, capsys):
    pattern_str = """
    [scopes.iterm]
    agent = "iterm"
    cursor = "profile"
    profile = "smoov"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    lines = out.splitlines()
    assert len(lines) == 2
    # profile should always be rendered first
    assert lines[1] == r'builtin echo -en "\e[0q"'


def test_cursor_style(dye_cmdline, capsys):
    pattern_str = """
    [scopes.iterm]
    agent = "iterm"
    styles.cursor = "#df769b"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, _ = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert out.count("\n") == 1
    assert "SetColors=curbg=df769b" in out



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

import os

import pytest
import rich

from dye import Dye, Pattern, Theme


@pytest.fixture
def dye_cmdline(mocker, monkeypatch):
    '''a fixture that simulates runing dye from the command line

    this fixture returns a function, which allows us to call
    the fixture and pass parameters to it

    def test_unset_list(dye_cmdline, capsys):
        theme = """
        [styles]
        text = "#dddddd on #222222
        """
        pattern = """
        [scope.ls]
        agent = "environment_variables"
        # set some environment variables
        unset = ["SOMEVAR", "ANOTHERVAR"]
        export.LS_COLORS = "ace ventura"
        """
        exit_code = dye_cmdline("apply -c", styles, pattern)
        ...

    This fixture allows us to pass a theme and a pattern in as strings,
    and also pass in any 'dye' command line we want. The fixture then
    patches the styles and pattern into 'dye' and runs the command line.

    Very convenient.
    '''
    # scrub environment variables that affect dye's behavior
    clean_env = {
        k: v for k, v in os.environ.items() if not k.startswith(("DYE_", "NO_COLOR"))
    }
    monkeypatch.setattr(os, "environ", clean_env)

    # patch up the console objects so we get ansi output and don't wrap text
    console_patch = mocker.patch("dye.Dye._create_console")
    console_patch.return_value = rich.console.Console(
        soft_wrap=True,
        markup=False,
        emoji=False,
        highlight=False,
        # force display to true color so we can look for ansi codes in output
        color_system="truecolor",
        # don't let it autodetect the width, we don't want any line wrapping
        width=2048,
    )

    error_console_patch = mocker.patch("dye.Dye._create_error_console")
    error_console_patch.return_value = rich.console.Console(
        stderr=True,
        soft_wrap=True,
        markup=False,
        emoji=False,
        highlight=False,
        # force display to true color so we can look for ansi codes in output
        color_system="truecolor",
        # don't let it autodetect the width, we don't want any line wrapping
        width=2048,
    )

    console_patch = mocker.patch("dye.Dye._create_print_console")
    console_patch.return_value = rich.console.Console(
        soft_wrap=True,
        markup=True,
        emoji=False,
        highlight=False,
        # force display to true color so we can look for ansi codes in output
        color_system="truecolor",
        # don't let it autodetect the width, we don't want any line wrapping
        width=2048,
    )

    def _executor(cmdline, theme_toml=None, pattern_toml=None):
        """anything that outputs here doesn't get captured by the
        capsys fixture"""
        if isinstance(cmdline, str):
            argv = cmdline.split(" ")
        elif isinstance(cmdline, list):
            argv = cmdline
        else:
            argv = []

        dye = Dye()
        try:
            (prog, args) = dye.parse_args(argv)
        except SystemExit as exc:
            return exc.code

        # create theme and pattern objects from the toml we were given
        theme = Theme.loads(theme_toml)
        pattern = Pattern.loads(pattern_toml, theme)

        # patch Dye methods to return our objects
        theme_patch = mocker.patch("dye.Dye.load_theme_from_args")
        theme_patch.return_value = theme
        pattern_patch = mocker.patch("dye.Dye.load_pattern_from_args")
        pattern_patch.return_value = pattern

        # now go run the command
        return dye.dispatch("dye", args)

    return _executor



#
# Copyright (c) 2025 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable
"""test all the jinja filters

test all the filters for functionality and that the list of all filters
has everything in it
"""

import pytest

from dye.filters import jinja_filters
from dye.pattern import Pattern

#
# jinja_filters()
#
ALL_FILTERS = [
    "fg_hex",
    "fg_hex_no_hash",
    "fg_rgb",
    "bg_hex",
    "bg_hex_no_hash",
    "bg_rgb",
    "ansi_on",
    "ansi_off",
]


def test_filter_list():
    # we don't care about the order, but we need to make
    # sure we have them all and no extras
    # ensuring we have no extras will prompt to write
    # more tests if a filter is added but there are no
    # tests for it
    f = list(jinja_filters().keys())
    assert f.sort() == ALL_FILTERS.sort()


TEMPLATES = [
    # fg_hex
    ("{{styles.dark_orange|fg_hex}}", "#ff6c1c"),
    ("{{styles.pink|fg_hex}}", "#df769b"),
    ("{{styles.cyan|fg_hex}}", "#09ecff"),
    ("{{styles.default|fg_hex}}", ""),
    ("{{styles.nothing|fg_hex}}", ""),
    ("{{variables.something|fg_hex}}", "Hello There."),
    # fg_hex_no_hash
    ("{{styles.dark_orange|fg_hex_no_hash}}", "ff6c1c"),
    ("{{styles.pink|fg_hex_no_hash}}", "df769b"),
    ("{{styles.cyan|fg_hex_no_hash}}", "09ecff"),
    ("{{styles.default|fg_hex_no_hash}}", ""),
    ("{{styles.nothing|fg_hex_no_hash}}", ""),
    ("{{variables.something|fg_hex_no_hash}}", "Hello There."),
    # fg_rgb
    ("{{styles.dark_orange|fg_rgb}}", "rgb(255,108,28)"),
    ("{{styles.pink|fg_rgb}}", "rgb(223,118,155)"),
    ("{{styles.cyan|fg_rgb}}", "rgb(9,236,255)"),
    ("{{styles.default|fg_rgb}}", ""),
    ("{{styles.nothing|fg_rgb}}", ""),
    ("{{variables.something|fg_rgb}}", "Hello There."),
    # bg_hex
    ("{{styles.dark_orange|bg_hex}}", "#222222"),
    ("{{styles.pink|bg_hex}}", ""),
    ("{{styles.cyan|bg_hex}}", ""),
    ("{{styles.default|bg_hex}}", ""),
    ("{{styles.nothing|bg_hex}}", ""),
    ("{{variables.something|bg_hex}}", "Hello There."),
    # bg_hex_no_hash
    ("{{styles.dark_orange|bg_hex_no_hash}}", "222222"),
    ("{{styles.pink|bg_hex_no_hash}}", ""),
    ("{{styles.cyan|bg_hex_no_hash}}", ""),
    ("{{styles.default|bg_hex_no_hash}}", ""),
    ("{{styles.nothing|bg_hex_no_hash}}", ""),
    ("{{variables.something|bg_hex_no_hash}}", "Hello There."),
    # bg_rgb
    ("{{styles.dark_orange|bg_rgb}}", "rgb(34,34,34)"),
    ("{{styles.pink|bg_rgb}}", ""),
    ("{{styles.cyan|bg_rgb}}", ""),
    ("{{styles.default|bg_rgb}}", ""),
    ("{{styles.nothing|bg_rgb}}", ""),
    ("{{variables.something|bg_rgb}}", "Hello There."),
    # ansi_on
    ("{{variables.something|ansi_on}}", "Hello There."),
    # ansi_off
    ("{{variables.something|ansi_off}}", "Hello There."),
]


@pytest.mark.parametrize("template, rendered", TEMPLATES)
def test_filters(template, rendered):
    pattern_str = f"""
        [styles]
        dark_orange = "#ff6c1c on #222222"
        pink = "bold #df769b"
        cyan = "#09ecff on default"
        default = "default"
        nothing = ""

        [variables]
        something = "Hello There."
        output = "echo {template}"
    """
    pattern = Pattern.loads(pattern_str)
    assert pattern.variables["output"] == f"echo {rendered}"


def test_ansi_on_off():
    """
    pattern_str has two kinds of embedded processing

    First, the python f-string takes the template argument
    and puts it where {template} is

    Second, jinja is going to process the whole string and pick up the
    {{ colors.background }} thing
    """
    pattern_str = """
        [styles]
        orng = "#ff6c1c on #222222"
        pink = "bold #df769b"
        cyan = "#09ecff on default"
        default = "default"
        nothing = ""

        [variables]
        something = "Hello There."
        opts = "--prompt={{styles.orng|ansi_on}}>>{{styles.orng|ansi_off}}"
    """
    pattern = Pattern.loads(pattern_str)
    expected_plaintext = 'export OPTS="--prompt=>>"\n'
    assert len(pattern.variables["opts"]) > len(expected_plaintext)



#
# Copyright (c) 2025 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

import pytest
import rich

import dye
from dye import Dye, Pattern

#
# test the fzf agent
#
ATTRIBS_TO_FZF = [
    ("bold", "regular:bold"),
    ("underline", "regular:underline"),
    ("reverse", "regular:reverse"),
    ("dim", "regular:dim"),
    ("italic", "regular:italic"),
    ("strike", "regular:strikethrough"),
    ("bold underline", "regular:bold:underline"),
    ("underline italic", "regular:underline:italic"),
    ("italic underline", "regular:underline:italic"),
]


@pytest.mark.parametrize("styledef, fzf", ATTRIBS_TO_FZF)
def test_fzf_attribs_from_style(styledef, fzf):
    style = rich.style.Style.parse(styledef)
    # we have to have a pattern in order for the agent to initialize
    # so lets make a fake one
    pattern_str = """
    [scopes.myscope]
    agent = "fzf"
    """
    pattern = Pattern.loads(pattern_str)
    agent = dye.agents.Fzf(pattern.scopes["myscope"])
    assert fzf == agent._fzf_attribs_from_style(style)


STYLE_TO_FZF = [
    # text, current-line, selected-line and preview styles have special processing
    # for foreground and background colors
    ("text", "", ""),
    ("text", "default", "fg:-1:regular"),
    ("text", "default on default", "fg:-1:regular,bg:-1"),
    ("text", "bold default on default underline", "fg:-1:regular:bold:underline,bg:-1"),
    ("text", "white on bright_red", "fg:7:regular,bg:9"),
    ("text", "bright_white", "fg:15:regular"),
    ("text", "bright_yellow on color(4)", "fg:11:regular,bg:4"),
    ("text", "green4", "fg:28:regular"),
    ("current-line", "navy_blue dim on grey82", "fg+:17:regular:dim,bg+:252"),
    (
        "selected-line",
        "navy_blue dim on grey82",
        "selected-fg:17:regular:dim,selected-bg:252",
    ),
    ("preview", "#af00ff on bright_white", "preview-fg:#af00ff:regular,preview-bg:15"),
    # other styles do not
    ("border", "magenta", "border:5:regular"),
    ("query", "#2932dc", "query:#2932dc:regular"),
]


@pytest.mark.parametrize("name, styledef, fzf", STYLE_TO_FZF)
def test_fzf_from_style(name, styledef, fzf):
    style = rich.style.Style.parse(styledef)
    # we have to have a pattern in order for the agent to initialize
    # so lets make a fake one
    pattern_str = """
    [scopes.myscope]
    agent = "fzf"
    """
    pattern = Pattern.loads(pattern_str)
    agent = dye.agents.Fzf(pattern.scopes["myscope"])
    assert fzf == agent._fzf_from_style(name, style)


def test_fzf(dye_cmdline, capsys):
    pattern_str = """
        [styles]
        purple = "#7060eb"

        [variables]
        bstyle = "rounded"

        [scopes.fzf]
        agent = "fzf"
        environment_variable = "QQQ"
        opts."+i" = true
        opts.--border = "{{variables.bstyle}}"
        styles.prompt = "magenta3"
        styles.info = "{{styles.purple}}"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    expected = (
        """export QQQ=" +i --border='rounded'"""
        """ --color='prompt:164:regular,info:#7060eb:regular'"\n"""
    )
    assert out == expected


def test_fzf_no_opts(dye_cmdline, capsys):
    pattern_str = """
        [scopes.fzf]
        agent = "fzf"
        environment_variable = "QQQ"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert out == """export QQQ=""\n"""


def test_fzf_no_varname(dye_cmdline, capsys):
    pattern_str = """
        [scopes.fzf]
        agent = "fzf"
        opts."+i" = true
        opts.--border = "rounded"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert "FZF_DEFAULT_OPTS" in out



#
# Copyright (c) 2025 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

import dye
from dye import Pattern


#
# test AgentBase
#
def test_agent_base_classmap():
    classmap = dye.agents.AgentBase.classmap
    assert "environment_variables" in classmap
    assert "bogusagent" not in classmap
    assert classmap["environment_variables"].__name__ == "EnvironmentVariables"


def test_agent_base_name():
    # we have to have a scope to initialze the agent
    # so lets make a fake one
    pattern_str = """
    [scopes.scope1]
    agent = "environment_variables"

    [scopes.scope2]
    agent = "fzf"
    """
    pattern = Pattern.loads(pattern_str)
    eza = dye.agents.EnvironmentVariables(pattern.scopes["scope1"])
    assert eza.agent_name == "environment_variables"
    fzfgen = dye.agents.Fzf(pattern.scopes["scope2"])
    assert fzfgen.agent_name == "fzf"



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

from dye import Dye


#
# test the shell agent
#
def test_shell(dye_cmdline, capsys):
    """
    pattern_str has two kinds of embedded processing

    First, the python f-string takes the template argument
    and puts it where {template} is

    Second, jinja is going to process the whole string and pick up the
    {{ colors.background }} thing

    These tests aren't comprehensive for template rendering, but we do need
    to make sure that it renders something, because it's up to the agent
    to call the template rendering code
    """
    pattern_str = """
            [colors]
            background = "#222222"

            [styles]
            dark_orange = "#ff6c1c on {{ colors.background }}"

            [variables]
            greeting = "Hello There."
            response = "General Kenobi."

            [scopes.obi-wan]
            agent = "shell"
            command.first = "echo {{variables.greeting}}"
            command.next = "printf {{variables.response}}"
            command.last = "echo {{styles.dark_orange|bg_hex}}"
        """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert out == "echo Hello There.\nprintf General Kenobi.\necho #222222\n"


def test_shell_multiline(dye_cmdline, capsys):
    pattern_str = """
        [variables]
        greeting = "hello there"

        [scopes.multiline]
        agent = "shell"
        command.long = '''
echo {{variables.greeting}}
echo general kenobi
if [[ 1 == 1 ]]; then
  echo "yes sir"
fi
'''
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    expected = """echo hello there
echo general kenobi
if [[ 1 == 1 ]]; then
  echo "yes sir"
fi
"""
    assert out == expected


def test_shell_no_commands(dye_cmdline, capsys):
    pattern_str = """
        [scopes.noop]
        agent = "shell"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert not out



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""the 'dye' command line tool for maintaining and switching color schemes"""

import argparse
import contextlib
import inspect
import os
import pathlib

import rich.box
import rich.color
import rich.console
import rich.errors
import rich.layout
import rich.rule
import rich.style
import rich.text
from benedict import benedict
from rich_argparse import RichHelpFormatter

from .agents import AgentBase
from .exceptions import DyeError, DyeSyntaxError
from .pattern import Pattern
from .theme import Theme
from .utils import version_string


class Dye:
    """parse and translate a theme file for various command line programs"""

    EXIT_SUCCESS = 0
    EXIT_ERROR = 1
    EXIT_USAGE = 2

    # we call these ELEMENTS instead of STYLES because we already use STYLES for
    # something else
    OUTPUT_ELEMENTS = [
        "usage_args",
        "usage_groups",
        "usage_help",
        "usage_metavar",
        "usage_prog",
        "usage_syntax",
        "usage_text",
        "ui_border",
        "ui_column_header",
        "error_progname",
        "error_text",
        "debug_label",
        "debug_text",
        "comment_begin",
        "comment_text",
    ]

    #
    # initialization and properties
    #
    def __init__(self):
        """Construct a new Dye object"""

        self.debug = False

        self.console = self._create_console(False)
        self.error_console = self._create_error_console(False)
        self.print_console = self._create_print_console(False)
        self.output_elements = benedict()

    @property
    def dye_dir(self):
        """Get the dye configuration directory from the shell environment

        returns a pathlib.Path if $DYE_DIR is set, else None
        """
        try:
            return pathlib.Path(os.environ["DYE_DIR"])
        except KeyError:
            return None

    @property
    def dye_theme_dir(self):
        """Get the dye themes directory

        returns a pathlib.Path if $DYE_DIR is set, else None
        """
        if self.dye_dir:
            return self.dye_dir / "themes"

        return None

    #
    # methods to process command line arguments and dispatch them
    # to the appropriate methods for execution
    #
    def dispatch(self, prog, args):
        """process and execute all the arguments and options"""

        # now go process everything (order matters)
        try:
            if args.help or args.command == "help":
                self.console.print(self.argparser().format_help())
                exit_code = self.EXIT_SUCCESS
            elif args.version:
                print(f"{prog} {version_string()}")
                exit_code = self.EXIT_SUCCESS
            elif not args.command:
                # this is a usage error, so it goes to stderr
                self.error_console.print(self.argparser().format_help())
                exit_code = self.EXIT_USAGE
            elif args.command == "apply":
                exit_code = self.command_apply(args)
            elif args.command == "preview":
                exit_code = self.command_preview(args)
            elif args.command == "print":
                exit_code = self.command_print(args)
            elif args.command == "agents":
                exit_code = self.command_agents(args)
            elif args.command == "themes":
                exit_code = self.command_themes(args)
            else:
                self.error_msg(prog, f"{args.command}: unknown command")
                exit_code = self.EXIT_USAGE
        except (DyeError, DyeSyntaxError) as err:
            self.error_msg(prog, str(err))
            exit_code = self.EXIT_ERROR

        return exit_code

    #
    # functions for the various commands called by dispatch()
    #
    def command_apply(self, args):
        """apply a pattern

        many agents just output to standard output, which we rely on a shell
        wrapper to execute for us. agents can also write/move files,
        replace files or whatever else they are gonna do

        output is suitable for `source <(dye apply)`
        """
        theme = self.load_theme_from_args(args, required=False)
        pattern = self.load_pattern_from_args(args, required=True, theme=theme)

        # if we got scope(s) on the command line, use them, otherwise we'll
        # apply all scopes
        to_apply = args.scope.split(",") if args.scope else list(pattern.scopes.keys())

        for scope_name in to_apply:
            # checking here in case they supplied a scope on the command line that
            # doesn't exist
            try:
                scope = pattern.scopes[scope_name]
            except KeyError as exc:
                raise DyeError(f"{scope_name}: no such scope") from exc
            scope.run_agent(args.comment, self.console, self.output_elements)
        return self.EXIT_SUCCESS

    def command_preview(self, args):
        """Display a preview of the styles in a theme"""
        theme = self.load_theme_from_args(args, required=False)
        pattern = self.load_pattern_from_args(args, required=False, theme=theme)
        if not theme and not pattern:
            raise DyeError("nothing to preview")

        # the text style here makes the whole panel print with the foreground
        # and background colors from the style
        try:
            text_style = pattern.styles["text"]
        except KeyError as exc:
            raise DyeSyntaxError("no 'text' style defined") from exc

        outer_table = rich.table.Table(
            box=rich.box.ROUNDED,
            show_lines=True,
            expand=True,
            show_header=False,
            padding=0,
        )

        # output some basic information about the theme and pattern
        summary_table = rich.table.Table(
            box=None,
            expand=False,
            show_header=False,
            padding=(0, 0, 0, 1),
        )

        if theme:
            summary_table.add_row("Theme file:", theme.filename)
            theme_description = (
                f' = "{theme.description}"' if theme.description else " ="
            )
            summary_table.add_row("  description", theme_description)
            theme_type = f' = "{theme.type}"' if theme.type else " ="
            summary_table.add_row("  type", theme_type)
            theme_version = f' = "{theme.version}"' if theme.version else " ="
            summary_table.add_row("  version", theme_version)
        else:
            summary_table.add_row("No theme file.")
        summary_table.add_row("")

        if pattern:
            summary_table.add_row("Pattern file:", pattern.filename)
            pattern_description = (
                f' = "{pattern.description}"' if pattern.description else " ="
            )
            summary_table.add_row("  description", pattern_description)
            pattern_type = f' = "{pattern.type}"' if pattern.type else " ="
            summary_table.add_row("  type", pattern_type)
            pattern_version = f' = "{pattern.version}"' if pattern.version else " ="
            summary_table.add_row("  version", pattern_version)
        else:
            summary_table.add_row("No pattern file.")

        # add the summary table to the outer table
        outer_table.add_row(" Inputs")
        outer_table.add_row(summary_table)

        # add colors
        colors_table = rich.table.Table(
            box=None,
            expand=False,
            padding=(0, 0, 0, 1),
            style=text_style,
        )
        outer_table.add_row(" [colors]")
        outer_table.add_section()
        for color_name in pattern.colors.keypaths(sort=False):
            if isinstance(pattern.colors[color_name], dict):
                # this one is a nested table, skip it
                continue
            (source, color_def) = pattern.get_color_def(color_name)
            resolved_color = pattern.colors[color_name]
            if resolved_color:
                # we think there is some definition that isn't empty
                col1 = rich.text.Text.assemble(("██", resolved_color), f" {color_name}")
            else:
                # no definition, don't try to display the color
                col1 = f" ∅ {color_name}"
            col2 = rich.text.Text(f' = "{color_def}"  # from {source}')
            colors_table.add_row(col1, col2)
        outer_table.add_row(colors_table)

        # add styles
        styles_table = rich.table.Table(
            box=None,
            expand=False,
            padding=(0, 0, 0, 1),
            style=text_style,
        )
        outer_table.add_row(" [styles]")
        outer_table.add_section()
        for style_name in pattern.styles.keypaths(sort=False):
            if isinstance(pattern.styles[style_name], dict):
                # this one is a nested table, skip it
                continue
            (source, style_def) = pattern.get_style_def(style_name)
            style = pattern.styles[style_name]
            if style:
                col1 = rich.text.Text(style_name, style)
            else:
                col1 = rich.text.Text(f"∅ {style_name}")
            col2 = rich.text.Text(f' = "{style_def}"  # from {source}')
            styles_table.add_row(col1, col2)
        outer_table.add_row(styles_table)

        # print the whole thing out
        self.console.print(
            rich.panel.Panel(outer_table, box=rich.box.SIMPLE, style=text_style)
        )
        return self.EXIT_SUCCESS

    def command_print(self, args):
        """print arbitrary strings applying styles from a theme or pattern"""
        theme = self.load_theme_from_args(args, required=False)
        pattern = self.load_pattern_from_args(args, required=False, theme=theme)

        style = None
        if args.style:
            with contextlib.suppress(KeyError):
                style = pattern.styles[args.style]

        # build a rich.theme object from our styles
        rich_theme = rich.theme.Theme(pattern.styles)
        # and use it for just a moment
        with self.print_console.use_theme(rich_theme, inherit=False):
            if args.n:
                self.print_console.print(" ".join(args.string), style=style, end="")
            else:
                self.print_console.print(" ".join(args.string), style=style)
        return self.EXIT_SUCCESS

    def command_agents(self, _):
        """list all available agents and a short description of each"""
        # ignore all other args
        agents = {}
        for name, clss in AgentBase.classmap.items():
            desc = inspect.getdoc(clss)
            if desc:
                desc = desc.split("\n", maxsplit=1)[0]
            agents[name] = desc

        border_style = None
        with contextlib.suppress(KeyError):
            border_style = self.output_elements["ui_border"]

        header_style = None
        with contextlib.suppress(KeyError):
            header_style = self.output_elements["ui_column_header"]

        table = rich.table.Table(
            box=rich.box.ROUNDED,
            border_style=border_style,
            header_style=header_style,
        )
        table.add_column("Agent")
        table.add_column("Description")

        for agent in sorted(agents):
            table.add_row(agent, agents[agent])

        # panel = rich.panel.Panel(
        # table,
        # border_style="yellow on blue",
        # )

        self.console.print(table)

        return self.EXIT_SUCCESS

    def command_themes(self, _):
        """Print a list of all themes"""
        # ignore all other args
        if not self.dye_theme_dir:
            errmsg = (
                "the DYE_DIR environment variable must be set"
                " and that directory must contain a 'themes' directory"
            )
            raise DyeError(errmsg)
        if not self.dye_theme_dir.is_dir():
            errmsg = f"{self.dye_theme_dir}: is not a directory"
            raise DyeError(errmsg)

        themeglob = self.dye_theme_dir.glob("*.toml")
        themes = []
        for theme in themeglob:
            themes.append(theme.stem)
        themes.sort()
        for theme in themes:
            print(theme)
        return self.EXIT_SUCCESS

    #
    # supporting methods
    #

    def _create_console(self, force_color):
        """create a rich console object to be used for output

        we have this as a separate method so that it can be patched
        in our test suite
        """
        # force_terminal can be True, False, or None
        # argparse will always set it to be True or False
        # we need it to be True or None
        if not force_color:
            force_color = None
        return rich.console.Console(
            soft_wrap=True,
            markup=False,
            emoji=False,
            highlight=False,
            force_terminal=force_color,
        )

    def _create_error_console(self, force_color):
        """create a rich console object to be used for std err

        we have this as a separate method so that it can be patched
        in our test suite
        """
        # force_terminal can be True, False, or None
        # argparse will always set it to be True or False
        # we need it to be True or None
        if not force_color:
            force_color = None
        return rich.console.Console(
            stderr=True,
            soft_wrap=True,
            markup=False,
            emoji=False,
            highlight=False,
            force_terminal=force_color,
        )

    def _create_print_console(self, force_color):
        """create a rich console object to be used for output for the
        print command, which enables console markup

        we have this as a separate method so that it can be patched
        in our test suite
        """
        # force_terminal can be True, False, or None
        # argparse will always set it to be True or False
        # we need it to be True or None
        if not force_color:
            force_color = None
        return rich.console.Console(
            soft_wrap=True,
            markup=True,
            emoji=False,
            highlight=False,
            force_terminal=force_color,
        )

    def error_msg(self, prog, msg):
        """print a styled error message to stderr"""
        text = rich.text.Text()
        text.append(
            f"{prog}: ",
            style=self.output_elements.get("error_progname"),
        )
        text.append(msg, style=self.output_elements.get("error_text"))
        self.error_console.print(text)

    def debug_msg(self, msg):
        """print the message to stderr if debug is enabled"""
        if self.debug:
            text = rich.text.Text()
            text.append(
                "[debug] ",
                style=self.output_elements.get("debug_label"),
            )
            text.append(
                msg,
                style=self.output_elements.get("debug_text"),
            )
            self.error_console.print(text)

    def set_output_elements(self):
        """set the color elements for all our output

        Use the contents of DYE_COLORS env variable

        DYE_COLORS=usage_args=red bold on black:usage_groups=white on red:
        """
        self.output_elements = benedict()
        oe = benedict()
        try:
            env_colors = os.environ["DYE_COLORS"]
            oe = self._parse_colorspec(env_colors)
            self.debug_msg("parsed environment variable DYE_COLORS")
        except KeyError:
            env_colors = None

        # https://no-color.org/
        try:
            env_no_color = os.environ["NO_COLOR"]
            if not env_no_color:
                self.debug_msg(
                    "ignoring NO_COLOR environment variable which is set, but empty"
                )
        except KeyError:
            env_no_color = None
            self.debug_msg("NO_COLOR environment variable is not set")

        if env_no_color:
            # help_styles = self._parse_colorspec("")
            self.debug_msg(
                "no color output because NO_COLOR environment variable is set"
            )
        elif env_colors:
            # was set, and was set to a non-empty string
            self.output_elements = oe
            self.debug_msg("output colors set from DYE_COLORS environment variable")
        elif env_colors == "":
            # None is different than an empty string
            # disable the default color output
            # help_styles = self._parse_colorspec("")
            self.debug_msg(
                "no color output because DYE_COLORS environment"
                " variable is an empty string"
            )
        else:
            self.debug_msg(
                "no color output because DYE_COLORS environment variable is not set"
            )

        # transfer the usage elements into RichHelpFormatter
        RichHelpFormatter.styles = {}
        for key, value in self.output_elements.items():
            if key.startswith("usage_"):
                argparse_key = key[len("usage_") :]
                RichHelpFormatter.styles[f"argparse.{argparse_key}"] = value

    def _parse_colorspec(self, colorspec):
        "parse colorspec into a benedict of elements and styles"
        colors = benedict()
        ### set everything to default, ie smash all the default colors
        # for element in self.OUTPUT_ELEMENTS:
        #    colors[element] = "default"

        clauses = colorspec.split(":")
        for clause in clauses:
            parts = clause.split("=", 1)
            if len(parts) == 2:
                element = parts[0]
                styledef = parts[1]
                if element in self.OUTPUT_ELEMENTS:
                    colors[element] = styledef
                else:
                    self.debug_msg(
                        f"skipping invalid element in DYE_COLORS: '{element}'"
                    )
            else:
                # invalid syntax, too many equals signs
                # ignore this clause
                self.debug_msg(f"skipping invalid expression in DYE_COLORS: '{clause}'")
        return colors

    def parse_args(self, argv=None):
        """parse argv and return prog and args

        also parses environment variables and sets:

        self.output_elements
        self.debug
        self.console
        self.error_console
        self.print_console
        """
        # we are doing a hack here because we have a chicken and egg problem
        # we want debug messaging generated from set_output_elements(), but
        # set_output_elements() has to run before parse_args() because it
        # changes the color output of argparser. So.....
        if isinstance(argv, list) and (("-d" in argv) or ("--debug" in argv)):
            # this will get set again later too, which is OK
            self.debug = True
        self.set_output_elements()

        argparser = self.argparser()
        args = argparser.parse_args(argv)

        self.debug = args.debug

        if args.force_color:
            self.console = self._create_console(True)
            self.error_console = self._create_error_console(True)
            self.print_console = self._create_print_console(True)

        return (argparser.prog, args)

    def load_theme_from_args(self, args, required=True):
        """Load a theme from the command line args

        required - whether we have to have a theme file or not
                if required=False an empty theme can be returned

        Will raise an exception if args specify a file and it
        doesn't exist or can't be opened

        Resolution order:
        1. --theme-file, -t from the command line
        2. $DYE_THEME_FILE environment variable

        This returns a theme object

        :raises: an exception if we can't find a theme file

        """
        if required and args.no_theme:
            raise DyeError("a theme is required and you specified --no-theme")

        if not required and args.no_theme:
            self.debug_msg("skipping theme because --no-theme was specified")
            return Theme()

        fname = None

        if args.theme_file:
            fname = args.theme_file
        else:
            with contextlib.suppress(KeyError):
                fname = os.environ["DYE_THEME_FILE"]
                self.debug_msg(f"found theme '{fname}' in $DYE_THEME_FILE")

        if fname:
            with open(os.path.expanduser(fname), "rb") as file:
                self.debug_msg(f"loading theme from '{fname}'")
                theme = Theme.load(file, filename=fname)
            return theme

        if required:
            raise DyeError("no theme specified")

        self.debug_msg("no theme specified")
        return Theme()

    def load_pattern_from_args(self, args, required=True, theme=None):
        """load a pattern file from the args

        Resolution order:
        1. --pattern-file -f from the command line
        2. $DYE_PATTERN_FILE environment variable

        This returns a pattern object

        :raises: an exception if we can't find a pattern file

        It's up to the caller to match up the required parameters and the
        args in the namespace. If you pass required=False, this will check for
        args.no_pattern and ignore the environment variable and args.pattern_file
        """
        # if we don't have a no_pattern attribute, make sure it's set to false
        # some versions of the args we call with (like those for apply), don't
        # have this argument because it doesn't make any sense
        if not hasattr(args, "no_pattern"):
            args.no_pattern = False

        if required and args.no_pattern:
            raise DyeError("a pattern is required and you specified --no-pattern")

        if not required and args.no_pattern:
            self.debug_msg("skipping pattern because --no-pattern was specified")
            return Pattern.loads(None, theme)

        fname = None

        if args.pattern_file:
            fname = args.pattern_file
        else:
            with contextlib.suppress(KeyError):
                fname = os.environ["DYE_PATTERN_FILE"]
                self.debug_msg(f"found pattern '{fname}' in DYE_PATTERN_FILE")

        if fname:
            with open(os.path.expanduser(fname), "rb") as fobj:
                self.debug_msg(f"loading pattern from '{fname}'")
                pattern = Pattern.load(fobj, theme, filename=fname)
            return pattern

        if required:
            raise DyeError("no pattern specified")

        self.debug_msg("no pattern specified")
        return Pattern()

    #
    # methods for running from the command line and parsing arguments
    #
    @staticmethod
    def main(argv=None):
        """Entry point from the command line

        parse arguments and call dispatch() for processing
        """
        dye = Dye()
        try:
            (prog, args) = dye.parse_args(argv)
        except SystemExit as exc:
            return exc.code

        return dye.dispatch(prog, args)

    def argparser(self):
        """Build the argument parser"""

        RichHelpFormatter.usage_markup = True
        RichHelpFormatter.group_name_formatter = str.lower

        parser = argparse.ArgumentParser(
            description=(
                "activate color output in shell commands using themes and patterns"
            ),
            formatter_class=RichHelpFormatter,
            add_help=False,
            epilog=(
                "type  '[argparse.prog]%(prog)s[/argparse.prog]"
                " [argparse.args]<command>[/argparse.args] -h' for command"
                " specific help"
            ),
        )

        hgroup = parser.add_mutually_exclusive_group()
        help_help = "show this help message and exit"
        hgroup.add_argument(
            "-h",
            "--help",
            action="store_true",
            help=help_help,
        )
        version_help = "show the program version and exit"
        hgroup.add_argument(
            "-v",
            "--version",
            action="store_true",
            help=version_help,
        )

        forcecolor_help = "force color output to files and pipes"
        parser.add_argument(
            "-F", "--force-color", action="store_true", help=forcecolor_help
        )

        # debug
        debug_help = "output debug and status information to stderr"
        parser.add_argument("-d", "--debug", action="store_true", help=debug_help)

        # set up for the sub commands
        subparsers = parser.add_subparsers(
            dest="command",
            title="arguments",
            metavar="<command>",
            required=False,
            help="command to perform, which must be one of the following:",
        )

        self._argparser_apply(subparsers)
        self._argparser_preview(subparsers)
        self._argparser_print(subparsers)
        self._argparser_agents(subparsers)
        self._argparser_themes(subparsers)
        self._argparser_help(subparsers)

        return parser

    def _argparser_apply(self, subparsers):
        # apply command
        cmdhelp = "apply a theme"
        parser = subparsers.add_parser(
            "apply",
            description=cmdhelp,
            formatter_class=RichHelpFormatter,
            help=cmdhelp,
        )
        pfile_help = "specify a file containing a pattern"
        parser.add_argument("-f", "--pattern-file", metavar="<path>", help=pfile_help)

        theme_group = parser.add_mutually_exclusive_group()
        tfile_help = "specify a file containing a theme"
        theme_group.add_argument(
            "-t", "--theme-file", metavar="<path>", help=tfile_help
        )
        no_theme_help = "don't load any theme, ignores DYE_THEME_FILE"
        theme_group.add_argument("--no-theme", action="store_true", help=no_theme_help)

        scope_help = "only apply the given scope"
        parser.add_argument("-s", "--scope", help=scope_help)
        comment_help = "add comments to the generated shell output"
        parser.add_argument("-c", "--comment", action="store_true", help=comment_help)

    def _argparser_preview(self, subparsers):
        """Add a subparser for the preview command"""
        cmd_help = "show a preview of the styles in a theme"
        parser = subparsers.add_parser(
            "preview",
            description=cmd_help,
            formatter_class=RichHelpFormatter,
            help=cmd_help,
        )
        pattern_group = parser.add_mutually_exclusive_group()
        pfile_help = "specify a file containing a pattern"
        pattern_group.add_argument(
            "-f", "--pattern-file", metavar="<path>", help=pfile_help
        )
        no_pattern_help = "don't load any pattern, ignores DYE_PATTERN_FILE"
        pattern_group.add_argument(
            "--no-pattern", action="store_true", help=no_pattern_help
        )

        theme_group = parser.add_mutually_exclusive_group()
        tfile_help = "specify a file containing a theme"
        theme_group.add_argument(
            "-t", "--theme-file", metavar="<path>", help=tfile_help
        )
        no_theme_help = "don't load any theme, ignores DYE_THEME_FILE"
        theme_group.add_argument("--no-theme", action="store_true", help=no_theme_help)

    def _argparser_print(self, subparsers):
        """Add a subparser for the print command"""
        desc = """
            Print text using styles from a theme or pattern.
            String arguments are printed, seperated by a space, and followed
            by a newline character, to standard output.
        """
        # desc = textwrap.dedent(desc)
        parser = subparsers.add_parser(
            "print",
            description=desc,
            formatter_class=RichHelpFormatter,
            help="print text using styles from a theme or pattern",
        )
        newline_help = "do not append a newline"
        parser.add_argument("-n", action="store_true", help=newline_help)

        pattern_group = parser.add_mutually_exclusive_group()
        pfile_help = "specify a file containing a pattern"
        pattern_group.add_argument(
            "-f", "--pattern-file", metavar="<path>", help=pfile_help
        )
        no_pattern_help = "don't load any pattern, ignores DYE_PATTERN_FILE"
        pattern_group.add_argument(
            "--no-pattern", action="store_true", help=no_pattern_help
        )

        theme_group = parser.add_mutually_exclusive_group()
        tfile_help = "specify a file containing a theme"
        theme_group.add_argument(
            "-t", "--theme-file", metavar="<path>", help=tfile_help
        )
        no_theme_help = "don't load any theme, ignores DYE_THEME_FILE"
        theme_group.add_argument("--no-theme", action="store_true", help=no_theme_help)

        style_help = "apply this style to the output"
        parser.add_argument("-s", "--style", metavar="<style>", help=style_help)

        string_help = "strings to be printed"
        parser.add_argument("string", nargs="*", help=string_help)

    def _argparser_agents(self, subparsers):
        """Add a subparser for the agents command"""
        # agents command
        agents_help = "list all known agents"
        subparsers.add_parser(
            "agents",
            description=agents_help,
            formatter_class=RichHelpFormatter,
            help=agents_help,
        )

    def _argparser_themes(self, subparsers):
        """Add a subparser for the themes command"""
        themes_help = "list available themes"
        subparsers.add_parser(
            "themes",
            description=themes_help,
            formatter_class=RichHelpFormatter,
            help=themes_help,
        )

    def _argparser_help(self, subparsers):
        """Add a subparser for the help command"""
        help_help = "display this usage message"
        subparsers.add_parser(
            "help",
            description=help_help,
            formatter_class=RichHelpFormatter,
            help=help_help,
        )



#
# Copyright (c) 2025 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

import pytest

from dye.exceptions import DyeError
from dye.pattern import Pattern
from dye.scope import Scope

SAMPLE_PATTERN = """
[styles]
orange = "#d5971a"
cyan = "#09ecff"
purple = "#7060eb"

triad.first = "purple"
triad.second = "cyan"
triad.third = "orange"

[scopes.iterm]
agent = "iterm"
cursor = "block"

[scopes.nocolor]
agent = "environment_variables"
export.NO_COLOR = "true"

[scopes.shell]
agent = "shell"
is_enabled = false
command.dontrun = "echo qqq"

[scopes.fzf]
agent = "fzf"
styles.file = "orange"
styles.directory = "{{ style.cyan }}"
style.border = "purple"
# purple should win
styles.prompt = "purple"
style.prompt = "orange"

[scopes.fzf2]
agent = "fzf"
environment_variable = "MY_SPECIAL_FZF_OPTS"
style.file = "triad.first"
style.directory = "{{ style.triad.second }}"
styles.prompt = "{{ styles.triad.third }}"
"""


@pytest.fixture
def spat():
    pattern = Pattern.loads(SAMPLE_PATTERN)
    return pattern


def test_init_scope_not_found(spat):
    with pytest.raises(DyeError):
        Scope("scopedoesntexist", spat)


def test_scope_no_agent():
    pattern_str = """
    [scopes.noagent]
    """
    with pytest.raises(DyeError):
        Pattern.loads(pattern_str)


def test_scope_unknown_agent():
    pattern_str = """
    [scopes.unknown]
    agent = "fredflintstone"
    """
    with pytest.raises(DyeError):
        Pattern.loads(pattern_str)


def test_scope_styles_lookup(spat):
    scope = spat.scopes["fzf"]
    assert scope.styles["file"] == spat.styles["orange"]
    assert scope.styles["directory"] == spat.styles["cyan"]


def test_scope_style(spat):
    # check that you can use
    # style.file = "#ffffff" and it will work just like styles.file = "#ffffff" does
    scope = spat.scopes["fzf"]
    assert scope.styles["border"] == spat.styles["purple"]


def test_scope_styles_overrides_style(spat):
    # check that if you have both
    # styles.prompt = "#333333"
    # style.prompt = "#ffffff"
    # you get #333333
    scope = spat.scopes["fzf"]
    assert scope.styles["prompt"] == spat.styles["purple"]


def test_scope_styles_subtable1(spat):
    scope = spat.scopes["fzf2"]
    assert scope.styles["file"] == spat.styles["purple"]


def test_scope_styles_subtable2(spat):
    scope = spat.scopes["fzf2"]
    assert scope.styles["directory"] == spat.styles["cyan"]


def test_scope_styles_subtable3(spat):
    scope = spat.scopes["fzf2"]
    assert scope.styles["prompt"] == spat.styles["orange"]



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""iTerm terminal emulator agent"""

import contextlib

from ..exceptions import DyeSyntaxError
from .base import AgentBase


class Iterm(AgentBase):
    "Send escape sequences to iTerm terminal emulator"

    def run(self, comments=False):
        """send escape sequences to iTerm to make it do stuff"""
        output = []

        # set the profile
        self._iterm_profile(output)

        # set the tab or window title color
        self._iterm_tab(output)

        # set foreground and background colors
        self._iterm_render_style(output, "foreground", "fg")
        self._iterm_render_style(output, "background", "bg")

        # set the cursor shape and color
        self._iterm_cursor(output)

        return "\n".join(output)

    def _iterm_profile(self, output):
        """add commands to output to tell iterm to change to a profile"""
        try:
            profile = self.scope.definition["profile"]
            cmd = r'builtin echo -en "\e]1337;'
            cmd += f"SetProfile={profile}"
            cmd += r'\a"'
            output.append(cmd)
        except KeyError:
            # no profile directive given
            pass

    def _iterm_tab(self, output):
        """add commands to output to change the tab or window title background color"""
        with contextlib.suppress(KeyError):
            style = self.scope.styles["tab"]
            if style.color.is_default:
                # set the command to change the tab color back to the default,
                # meaning whatever is set in the profile.
                # gotta use raw strings here so the \e and \a don't get
                # interpreted by python, they need to be passed through
                # to the echo command
                cmd = r'builtin echo -en "\e]6;1;bg;*;default\a"'
                output.append(cmd)
            else:
                clr = style.color.get_truecolor()
                # in iterm you have to send different escape sequences
                #
                # gotta use raw strings here so the \e and \a don't get
                # interpreted by python, they need to be passed through
                # to the echo command
                cmd = r'builtin echo -en "\e]6;1;bg;red;brightness;'
                cmd += f"{clr.red}"
                cmd += r'\a"'
                output.append(cmd)
                cmd = r'builtin echo -en "\e]6;1;bg;green;brightness;'
                cmd += f"{clr.green}"
                cmd += r'\a"'
                output.append(cmd)
                cmd = r'builtin echo -en "\e]6;1;bg;blue;brightness;'
                cmd += f"{clr.blue}"
                cmd += r'\a"'
                output.append(cmd)

    CURSOR_MAP = {
        "block": "0",
        "box": "0",
        "vertical_bar": "1",
        "vertical": "1",
        "bar": "1",
        "pipe": "1",
        "underline": "2",
    }

    def _iterm_cursor(self, output):
        """create echo commands to change the cursor shape,
        foreground, and background colors
        """
        # check the cursor shape
        try:
            cursor = self.scope.definition["cursor"]
        except KeyError:
            cursor = None
        if cursor:
            if cursor == "profile":
                cmd = r'builtin echo -en "\e[0q"'
                output.append(cmd)
            elif cursor in self.CURSOR_MAP:
                cmd = r'builtin echo -en "\e]1337;'
                cmd += f"CursorShape={self.CURSOR_MAP[cursor]}"
                cmd += r'\a"'
                output.append(cmd)
            else:
                raise DyeSyntaxError(
                    f"unknown cursor '{cursor}' while processing scope '{self.scope}'"
                )
        # render the cursor color
        # iterm has curbg and curfg color codes, but as far as I can tell
        # the curfg is a noop
        self._iterm_render_style(output, "cursor", "curbg")

    def _iterm_render_style(self, output, style_name, iterm_key):
        """append an iterm escape sequence to change the color palette to output"""
        try:
            style = self.scope.styles[style_name]
            clr = style.color.get_truecolor()
            # gotta use raw strings here so the \e and \a don't get
            # interpreted by python, they need to be passed through
            # to the echo command
            cmd = r'builtin echo -en "\e]1337;'
            cmd += f"SetColors={iterm_key}={clr.hex.replace('#', '')}"
            cmd += r'\a"'
            output.append(cmd)
        except KeyError:
            # the given style doesn't exist
            pass



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""fzf agent for FZF_DEFAULT_OPTS"""

import rich.color

from .base import AgentBase


class Fzf(AgentBase):
    "Set fzf options and environment variables"

    def run(self, comments=False) -> str:
        """render attribs into a shell statement to set an environment variable"""
        optstr = ""
        # process all the command line options
        try:
            opts = self.scope.definition["opts"]
        except KeyError:
            opts = {}

        for key, value in opts.items():
            if isinstance(value, str):
                optstr += f" {key}='{value}'"
            elif isinstance(value, bool) and value:
                optstr += f" {key}"

        # process all the styles
        colors = []
        # then add them back
        for name, style in self.scope.styles.items():
            colors.append(self._fzf_from_style(name, style))
        # turn off all the colors, and add our color strings
        try:
            colorbase = f"{self.scope.definition['colorbase']},"
        except KeyError:
            colorbase = ""
        if colorbase or colors:
            colorstr = f" --color='{colorbase}{','.join(colors)}'"
        else:
            colorstr = ""

        # figure out which environment variable to put it in
        try:
            varname = self.scope.definition["environment_variable"]
        except KeyError:
            varname = "FZF_DEFAULT_OPTS"
        print(f'export {varname}="{optstr}{colorstr}"')

    def _fzf_from_style(self, name, style):
        """turn a rich.style into a valid fzf color"""
        # fzf has different color names for foreground and background items
        # we combine them
        name_map = {
            "text": ("fg", "bg"),
            "current-line": ("fg+", "bg+"),
            "selected-line": ("selected-fg", "selected-bg"),
            "preview": ("preview-fg", "preview-bg"),
        }

        fzf_colors = []
        if name in name_map:
            fgname, bgname = name_map[name]
            if style.color:
                fzfc = self._fzf_color_from_rich_color(style.color)
                fzfa = self._fzf_attribs_from_style(style)
                fzf_colors.append(f"{fgname}:{fzfc}:{fzfa}")
            if style.bgcolor:
                fzfc = self._fzf_color_from_rich_color(style.bgcolor)
                fzf_colors.append(f"{bgname}:{fzfc}")
        else:
            # we only use the foreground color of the style, and ignore
            # any background color specified by the style
            if style.color:
                fzfc = self._fzf_color_from_rich_color(style.color)
                fzfa = self._fzf_attribs_from_style(style)
                fzf_colors.append(f"{name}:{fzfc}:{fzfa}")

        return ",".join(fzf_colors)

    def _fzf_color_from_rich_color(self, color):
        """turn a rich.color into it's fzf equivilent"""
        fzf = ""

        if color.type == rich.color.ColorType.DEFAULT:
            fzf = "-1"
        elif color.type == rich.color.ColorType.STANDARD:
            # python rich uses underscores, fzf uses dashes
            fzf = str(color.number)
        elif color.type == rich.color.ColorType.EIGHT_BIT:
            fzf = str(color.number)
        elif color.type == rich.color.ColorType.TRUECOLOR:
            fzf = color.triplet.hex
        return fzf

    def _fzf_attribs_from_style(self, style):
        attribs = "regular"
        if style.bold:
            attribs += ":bold"
        if style.underline:
            attribs += ":underline"
        if style.reverse:
            attribs += ":reverse"
        if style.dim:
            attribs += ":dim"
        if style.italic:
            attribs += ":italic"
        if style.strike:
            attribs += ":strikethrough"
        return attribs



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

import os

import pytest
import rich
import rich.text
from rich_argparse import RichHelpFormatter

from dye import Dye, DyeError
from dye import __main__ as mainmodule


#
# test output color logic
#
def test_output_color_no_color_env(dye_cmdline, mocker):
    RichHelpFormatter.styles["argparse.text"] = "#ff00ff"
    mocker.patch.dict(os.environ, {"NO_COLOR": "doesn't matter"}, clear=True)
    dye_cmdline("--help")
    assert not RichHelpFormatter.styles


def test_output_color_envs_only(dye_cmdline, mocker):
    # NO_COLOR should override DYE_COLORS
    RichHelpFormatter.styles["argparse.text"] = "#333333"
    envs = {"DYE_COLORS": "usage_text=#f0f0f0", "NO_COLOR": "doesn't matter"}
    mocker.patch.dict(os.environ, envs, clear=True)
    dye_cmdline("--help")
    assert not RichHelpFormatter.styles


def test_output_color_no_color_env_empty(dye_cmdline, mocker, capsys):
    # NO_COLOR set but empty should be ignored per no-color.org spec
    mocker.patch.dict(os.environ, {"NO_COLOR": ""}, clear=True)
    exit_code = dye_cmdline("-d --help")
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert "ignoring NO_COLOR" in err


def test_output_color_env_color(dye_cmdline, mocker):
    # DYE_COLORS should override default colors
    RichHelpFormatter.styles["argparse.text"] = "#333333"
    mocker.patch.dict(os.environ, {"DYE_COLORS": "usage_text=#f0f0f0"}, clear=True)
    dye_cmdline("--help")
    assert RichHelpFormatter.styles["argparse.text"] == "#f0f0f0"


def test_output_color_force(mocker):
    create_console = mocker.patch("dye.Dye._create_console")
    create_error_console = mocker.patch("dye.Dye._create_error_console")
    create_print_console = mocker.patch("dye.Dye._create_print_console")
    dye = Dye()
    # reset so we only track calls from parse_args, not the constructor
    create_console.reset_mock()
    create_error_console.reset_mock()
    create_print_console.reset_mock()
    dye.parse_args(["--force-color", "agents"])
    create_console.assert_called_once_with(True)
    create_error_console.assert_called_once_with(True)
    create_print_console.assert_called_once_with(True)


def test_output_color_env_empty(dye_cmdline, mocker):
    # DYE_COLORS should override default colors
    RichHelpFormatter.styles["argparse.text"] = "#ff00ff"
    mocker.patch.dict(os.environ, {"DYE_COLORS": ""}, clear=True)
    dye_cmdline("--help")
    assert not RichHelpFormatter.styles


def test_output_color_env_unknown_element(dye_cmdline, mocker, capsys):
    mocker.patch.dict(
        os.environ,
        {"DYE_COLORS": "homer_simpson=red bold on black:fred=white"},
    )
    msg = "It's Wednesday my dudes."
    exit_code = dye_cmdline(f"-d print {msg}")
    out, err = capsys.readouterr()
    assert "[debug]" in err
    assert "skipping invalid element" in err
    assert msg in out
    assert exit_code == 0


def test_output_color_env_parse_error(dye_cmdline, mocker, capsys):
    mocker.patch.dict(
        os.environ,
        {"DYE_COLORS": "usage_help=red:usage_args red bold on black white"},
    )
    msg = "It's Wednesday my dudes."
    exit_code = dye_cmdline(f"-d print {msg}")
    out, err = capsys.readouterr()
    assert "[debug]" in err
    assert "skipping invalid expression" in err
    assert msg in out
    assert exit_code == 0


#
# test unknown commands, no commands, help, version, and debug
#
def test_help_option(dye_cmdline, capsys):
    exit_code = dye_cmdline("--help")
    assert exit_code == Dye.EXIT_SUCCESS
    out, err = capsys.readouterr()
    assert not err
    assert "preview" in out
    assert "--force-color" in out


def test_h_option(dye_cmdline, capsys):
    exit_code = dye_cmdline("-h")
    assert exit_code == Dye.EXIT_SUCCESS
    out, err = capsys.readouterr()
    assert not err
    assert "preview" in out
    assert "--force-color" in out


def test_version_option(dye_cmdline, capsys):
    exit_code = dye_cmdline("--version")
    assert exit_code == Dye.EXIT_SUCCESS
    out, err = capsys.readouterr()
    assert not err
    assert "dye" in out


def test_v_option(dye_cmdline, capsys):
    exit_code = dye_cmdline("-v")
    assert exit_code == Dye.EXIT_SUCCESS
    out, err = capsys.readouterr()
    assert not err
    assert "dye" in out


def test_h_and_v_option(dye_cmdline, capsys):
    exit_code = dye_cmdline("-h -v")
    assert exit_code == Dye.EXIT_USAGE
    out, err = capsys.readouterr()
    assert not out
    # this message comes from argparse, we can't modify it
    assert "not allowed with argument" in err


def test_no_command(dye_cmdline, capsys):
    # this should show the usage message
    exit_code = dye_cmdline(None)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_USAGE
    assert not out
    # if you don't give a command, that's a usage error
    # so the usage message goes on standard error
    # check a few things in the usage message
    assert "apply" in err
    assert "preview" in err
    assert "--force-color" in err
    assert "-v" in err


def test_help_command(dye_cmdline, capsys):
    # this should show the usage message
    exit_code = dye_cmdline("help")
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    # if you ask for help, help should be on standard output
    assert "apply" in out
    assert "preview" in out
    assert "--force-color" in out
    assert "-v" in out


def test_unknown_command(dye_cmdline, capsys):
    # these errors are all raised and generated by argparse
    exit_code = dye_cmdline("unknowncommand")
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_USAGE
    assert not out
    assert "error" in err
    assert "invalid choice" in err


def test_debug(capsys):
    # we don't use the dye_cmdline fixture because the debug output
    # is generated during dispatch, before the fixture can capture it;
    # we use --no-theme and --no-pattern to make sure the test isn't
    # reading any theme/pattern files the developer has set in their
    # environment
    msg = "Hello there. General Kenobi."
    dye = Dye()
    argv = ["-d", "print", "--no-theme", "--no-pattern", msg]
    (_, args) = dye.parse_args(argv)
    exit_code = dye.dispatch("dye", args)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert "[debug]" in err
    assert out == f"{msg}\n"


#
# test dye_dir() property
#
def test_dye_dir_environment_variable(mocker, tmp_path):
    dye = Dye()
    mocker.patch.dict(os.environ, {"DYE_DIR": str(tmp_path)})
    # dye_dir should be a Path object
    assert dye.dye_dir == tmp_path


def test_dye_dir_no_environment_variable(mocker):
    # ensure no DYE_DIR environment variable exists
    mocker.patch.dict(os.environ, {}, clear=True)
    dye = Dye()
    assert not dye.dye_dir


def test_dye_dir_invalid_directory(mocker, tmp_path):
    invalid = tmp_path / "doesntexist"
    mocker.patch.dict(os.environ, {"DYE_DIR": str(invalid)})
    dye = Dye()
    assert dye.dye_dir == invalid


#
# test Dye.main(), the entry point for the command line script
#
def test_dye_main(mocker):
    # we are just testing main() here, as long as it dispatches, we don't
    # care what the dispatch_list() function returns in this test
    dmock = mocker.patch("dye.Dye.command_agents")
    dmock.return_value = Dye.EXIT_SUCCESS
    assert Dye.main(["agents"]) == Dye.EXIT_SUCCESS


def test_dye_main_unknown_command():
    assert Dye.main(["unknowncommand"]) == Dye.EXIT_USAGE


def test_dispatch_unknown_command(capsys):
    # but by calling dispatch() directly, we can get our own errors
    # first we have to parse valid args
    dye = Dye()
    parser = dye.argparser()
    args = parser.parse_args(["agents"])
    # and then substitute a fake command
    args.command = "fredflintstone"
    exit_code = dye.dispatch("dye", args)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_USAGE
    assert not out
    assert "unknown command" in err


def test_error_styled_unknown_command(mocker, capsys):
    create_error_console = mocker.patch("dye.Dye._create_error_console")
    create_error_console.return_value = rich.console.Console(
        stderr=True,
        soft_wrap=True,
        markup=False,
        emoji=False,
        highlight=False,
        color_system="truecolor",
        width=2048,
    )
    mocker.patch.dict(
        os.environ,
        {"DYE_COLORS": "error_progname=red:error_text=#00ff00"},
        clear=True,
    )
    dye = Dye()
    argv = ["agents"]
    (_, args) = dye.parse_args(argv)
    args.command = "fredflintstone"
    exit_code = dye.dispatch("dye", args)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_USAGE
    assert not out
    assert "unknown command" in err
    # ANSI escape codes should be present because styles are applied
    assert "\x1b[" in err


def test_error_styled_exception(mocker, capsys):
    create_error_console = mocker.patch("dye.Dye._create_error_console")
    create_error_console.return_value = rich.console.Console(
        stderr=True,
        soft_wrap=True,
        markup=False,
        emoji=False,
        highlight=False,
        color_system="truecolor",
        width=2048,
    )
    mocker.patch.dict(
        os.environ,
        {"DYE_COLORS": "error_progname=red:error_text=#00ff00"},
        clear=True,
    )
    dye = Dye()
    argv = ["agents"]
    (_, args) = dye.parse_args(argv)
    # mock command_agents to raise a DyeError
    mocker.patch.object(dye, "command_agents", side_effect=DyeError("test error"))
    exit_code = dye.dispatch("dye", args)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_ERROR
    assert not out
    assert "test error" in err
    assert "dye: " in err
    # ANSI escape codes should be present because styles are applied
    assert "\x1b[" in err


def test_debug_styled_output(mocker, capsys):
    create_error_console = mocker.patch("dye.Dye._create_error_console")
    create_error_console.return_value = rich.console.Console(
        stderr=True,
        soft_wrap=True,
        markup=False,
        emoji=False,
        highlight=False,
        color_system="truecolor",
        width=2048,
    )
    mocker.patch.dict(
        os.environ,
        {"DYE_COLORS": "debug_label=red:debug_text=#00ff00"},
        clear=True,
    )
    dye = Dye()
    msg = "Hello there. General Kenobi."
    argv = ["-d", "print", "--no-theme", "--no-pattern", msg]
    (_, args) = dye.parse_args(argv)
    exit_code = dye.dispatch("dye", args)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert "[debug]" in err
    assert msg in out
    # ANSI escape codes should be present because styles are applied
    assert "\x1b[" in err


def test___main__(mocker):
    mocker.patch("dye.Dye.main", return_value=42)
    mocker.patch.object(mainmodule, "__name__", "__main__")
    with pytest.raises(SystemExit) as excinfo:
        mainmodule.bootstrap()
    # unpack the exception to see if got the return value
    assert excinfo.value.code == 42


#
# test all the variations of load_theme_from_args()
#
# I gave up trying to descriptively name these tests, there are too
# many combinations. The test names would have been 90 characters long.
# at the top of each test is a comment describing the combination this
# test validates
LOAD_THEME1_ARGVS = [
    "apply",
    "preview",
]


@pytest.mark.parametrize("argv", LOAD_THEME1_ARGVS)
def test_load_theme_from_args_apply1(argv, mocker):
    # no environment
    # argv = "apply"
    # required = True and False
    mocker.patch.dict(os.environ, {}, clear=True)
    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    with pytest.raises(DyeError):
        dye.load_theme_from_args(args, required=True)

    theme = dye.load_theme_from_args(args, required=False)
    assert not theme


def test_load_theme_from_args_apply2(mocker):
    # no environment
    # argv = "apply --no-theme"
    # required = False
    argv = "apply --no-theme"
    mocker.patch.dict(os.environ, {}, clear=True)
    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    with pytest.raises(DyeError):
        dye.load_theme_from_args(args, required=True)

    theme = dye.load_theme_from_args(args, required=False)
    assert theme.colors == {}
    assert theme.styles == {}


def test_load_theme_from_args_apply3(mocker, tmp_path):
    # no environment
    # argv = "apply --theme-file {exists}"
    # required = True and False

    mocker.patch.dict(os.environ, {}, clear=True)

    # go write a theme file that we can actually open
    themefile = tmp_path / "sometheme.toml"
    toml = """
    [styles]
    text = "#ffcc00 on #003322"
    """
    with open(themefile, "w", encoding="utf8") as fvar:
        fvar.write(toml)

    argv = f"apply --theme-file {themefile}"

    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    theme = dye.load_theme_from_args(args, required=True)
    assert isinstance(theme.styles["text"], rich.style.Style)

    theme = dye.load_theme_from_args(args, required=False)
    assert isinstance(theme.styles["text"], rich.style.Style)


def test_load_theme_from_args_apply4(mocker, tmp_path):
    # no environment
    # argv = "apply --theme-file {doesn't exist}"
    # required = True and False

    mocker.patch.dict(os.environ, {}, clear=True)

    # a theme file that doesn't exist
    themefile = tmp_path / "doesntexisttheme.toml"

    argv = f"apply --theme-file {themefile}"

    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    with pytest.raises(FileNotFoundError):
        dye.load_theme_from_args(args, required=True)
    with pytest.raises(FileNotFoundError):
        dye.load_theme_from_args(args, required=False)


def test_load_theme_from_args_apply5(mocker, tmp_path):
    # DYE_THEME_FILE exists
    # argv = "apply"
    # required = True and False

    # go write a theme file that we can actually open
    themefile = tmp_path / "sometheme.toml"
    toml = """
    [styles]
    text = "#ffcc00 on #003322"
    """
    with open(themefile, "w", encoding="utf8") as fvar:
        fvar.write(toml)

    mocker.patch.dict(os.environ, {"DYE_THEME_FILE": f"{themefile}"}, clear=True)

    argv = "apply"
    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    theme = dye.load_theme_from_args(args, required=True)
    assert isinstance(theme.styles["text"], rich.style.Style)
    theme = dye.load_theme_from_args(args, required=False)
    assert isinstance(theme.styles["text"], rich.style.Style)


def test_load_theme_from_args_apply6(mocker, tmp_path):
    # DYE_THEME_FILE exists
    # argv = "apply --theme-file {exists}"
    # required = True and False

    # go write a theme file that we can actually open
    envfile = tmp_path / "sometheme.toml"
    toml = """
    [styles]
    text = "#ffcc00 on #003322"
    """
    with open(envfile, "w", encoding="utf8") as fvar:
        fvar.write(toml)

    mocker.patch.dict(os.environ, {"DYE_THEME_FILE": f"{envfile}"}, clear=True)

    cmdfile = tmp_path / "othertheme.toml"
    toml = """
    [styles]
    current_line = "#ffcc00 on #003322"
    """
    with open(cmdfile, "w", encoding="utf8") as fvar:
        fvar.write(toml)

    argv = f"apply --theme-file {cmdfile}"
    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    theme = dye.load_theme_from_args(args, required=True)
    assert isinstance(theme.styles["current_line"], rich.style.Style)
    assert "text" not in theme.styles

    theme = dye.load_theme_from_args(args, required=False)
    assert isinstance(theme.styles["current_line"], rich.style.Style)
    assert "text" not in theme.styles


def test_load_theme_from_args_apply7(mocker, tmp_path):
    # DYE_THEME_FILE exists
    # argv = "apply --no-theme"
    # required = True and False

    # go write a theme file that we can actually open
    themefile = tmp_path / "sometheme.toml"
    toml = """
    [styles]
    text = "#ffcc00 on #003322"
    """
    with open(themefile, "w", encoding="utf8") as fvar:
        fvar.write(toml)

    mocker.patch.dict(os.environ, {"DYE_THEME_FILE": f"{themefile}"}, clear=True)

    argv = "apply --no-theme"
    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    with pytest.raises(DyeError):
        dye.load_theme_from_args(args, required=True)
    theme = dye.load_theme_from_args(args, required=False)
    assert theme.styles == {}


#
# test all the variations of load_pattern_from_args()
#
# I gave up trying to descriptively name these tests, there are too
# many combinations. The test names would have been 90 characters long.
# at the top of each test is a comment describing the combination this
# test validates
def test_load_pattern_from_args1(mocker):
    # no environment
    # argv = "print"
    # required = True and False
    argv = "print"
    mocker.patch.dict(os.environ, {}, clear=True)
    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    with pytest.raises(DyeError):
        dye.load_pattern_from_args(args, required=True)

    pattern = dye.load_pattern_from_args(args, required=False)
    assert pattern.colors == {}
    assert pattern.styles == {}


def test_load_pattern_from_args2(mocker):
    # no environment
    # argv = "print --no-pattern"
    # required = False
    argv = "print --no-pattern"
    mocker.patch.dict(os.environ, {}, clear=True)
    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    with pytest.raises(DyeError):
        dye.load_pattern_from_args(args, required=True)

    pattern = dye.load_pattern_from_args(args, required=False)
    assert pattern.colors == {}
    assert pattern.styles == {}


def test_load_pattern_from_args3(mocker, tmp_path):
    # no environment
    # argv = "print --pattern-file {exists}"
    # required = True and False

    mocker.patch.dict(os.environ, {}, clear=True)

    pattern_file = tmp_path / "pattern.toml"
    toml = """
    [styles]
    text = "#ffcc00 on #003322"
    """
    with open(pattern_file, "w", encoding="utf8") as fvar:
        fvar.write(toml)

    argv = f"print --pattern-file {pattern_file}"
    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    theme = dye.load_pattern_from_args(args, required=True)
    assert isinstance(theme.styles["text"], rich.style.Style)

    theme = dye.load_pattern_from_args(args, required=False)
    assert isinstance(theme.styles["text"], rich.style.Style)


def test_load_pattern_from_args4(mocker, tmp_path):
    # no environment
    # argv = "print --pattern-file {doesn't exist}"
    # required = True and False

    mocker.patch.dict(os.environ, {}, clear=True)

    pattern_file = tmp_path / "doesntexistpattern.toml"

    argv = f"print --pattern-file {pattern_file} something"
    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    with pytest.raises(FileNotFoundError):
        dye.load_pattern_from_args(args, required=True)
    with pytest.raises(FileNotFoundError):
        dye.load_pattern_from_args(args, required=False)


def test_load_pattern_from_args5(mocker, tmp_path):
    # DYE_PATTERN_FILE exists
    # argv = "print"
    # required = True and False

    pattern_file = tmp_path / "pattern.toml"
    toml = """
    [styles]
    text = "#ffcc00 on #003322"
    """
    with open(pattern_file, "w", encoding="utf8") as fvar:
        fvar.write(toml)

    mocker.patch.dict(os.environ, {"DYE_PATTERN_FILE": f"{pattern_file}"}, clear=True)

    argv = "apply"
    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    theme = dye.load_pattern_from_args(args, required=True)
    assert isinstance(theme.styles["text"], rich.style.Style)
    theme = dye.load_pattern_from_args(args, required=False)
    assert isinstance(theme.styles["text"], rich.style.Style)


def test_load_pattern_from_args6(mocker, tmp_path):
    # DYE_PATTERN_FILE exists
    # argv = "apply --pattern-file {exists}"
    # required = True and False

    envfile = tmp_path / "envpattern.toml"
    toml = """
    [styles]
    text = "#ffcc00 on #003322"
    """
    with open(envfile, "w", encoding="utf8") as fvar:
        fvar.write(toml)

    mocker.patch.dict(os.environ, {"DYE_PATTERN_FILE": f"{envfile}"}, clear=True)

    cmdfile = tmp_path / "cmdpattern.toml"
    toml = """
    [styles]
    current_line = "#ffcc00 on #003322"
    """
    with open(cmdfile, "w", encoding="utf8") as fvar:
        fvar.write(toml)

    argv = f"print --pattern-file {cmdfile}"
    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    theme = dye.load_pattern_from_args(args, required=True)
    assert isinstance(theme.styles["current_line"], rich.style.Style)
    assert "text" not in theme.styles

    theme = dye.load_pattern_from_args(args, required=False)
    assert isinstance(theme.styles["current_line"], rich.style.Style)
    assert "text" not in theme.styles


def test_load_pattern_from_args7(mocker, tmp_path):
    # DYE_PATTERN_FILE exists
    # argv = "print --no-pattern"
    # required = True and False

    pattern_file = tmp_path / "sometheme.toml"
    toml = """
    [styles]
    text = "#ffcc00 on #003322"
    """
    with open(pattern_file, "w", encoding="utf8") as fvar:
        fvar.write(toml)

    mocker.patch.dict(os.environ, {"DYE_PATTERN_FILE": f"{pattern_file}"}, clear=True)

    argv = "print --no-pattern"
    dye = Dye()
    (_, args) = dye.parse_args(argv.split())

    with pytest.raises(DyeError):
        dye.load_pattern_from_args(args, required=True)
    theme = dye.load_pattern_from_args(args, required=False)
    assert theme.styles == {}



#
# Copyright (c) 2025 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

from dye import Dye


#
# test the print command
#
def test_print_basic(dye_cmdline, capsys):
    strings = "Hello there. General Kenobi."
    exit_code = dye_cmdline(f"print {strings}")
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert out == f"{strings}\n"


def test_print_no_newline(dye_cmdline, capsys):
    strings = "Hello there. General Kenobi."
    exit_code = dye_cmdline(f"print -n {strings}")
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert out == f"{strings}"


def test_print_theme_style(dye_cmdline, capsys):
    theme_str = """
    [styles]
    bright_green = "#99e343 bold"
    """
    strings = "Hello there. General Kenobi."
    exit_code = dye_cmdline(f"print -s bright_green {strings}", theme_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert strings in out
    # there better be some ansi codes in there, but we aren't going to test which ones
    assert len(out) > len(f"{strings}\n")


def test_print_theme_invalid_style(dye_cmdline, capsys):
    theme_str = """
    [styles]
    bright_green = "#99e343 bold"
    """
    strings = "Hello there. General Kenobi."
    exit_code = dye_cmdline(f"print -s bbright_green {strings}", theme_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert out == f"{strings}\n"



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

import pytest

from dye import Dye

#
# test the environment_variables agent
#
TEMPLATES = [
    ("{{variables.someopts}}", "--option=fred -v"),
    ("{{styles.dark_orange|fg_hex}}", "#ff6c1c"),
]


@pytest.mark.parametrize("template, rendered", TEMPLATES)
def test_export(dye_cmdline, capsys, template, rendered):
    """
    pattern_str has two kinds of embedded processing

    First, the python f-string takes the template argument
    and puts it where {template} is

    Second, jinja is going to process the whole string and pick up the
    {{ colors.background }} thing

    These tests aren't comprehensive for template rendering, but we do need
    to make sure that it renders something, because it's up to the agent
    to call the template rendering code
    """
    pattern_str = (
        """
            [colors]
            background = "#222222"

            [styles]
            dark_orange = "#ff6c1c on {{ colors.background }}"

            [variables]
            someopts = "--option=fred -v"
        """
        f"""
            [scopes.gum]
            agent = "environment_variables"
            export.GUM_OPTS = " --cursor-foreground={template}"
            export.FRED = "{template}"
        """
    )
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    lines = out.splitlines()
    assert lines[0] == f'export GUM_OPTS=" --cursor-foreground={rendered}"'
    assert lines[1] == f'export FRED="{rendered}"'


def test_unset_list(dye_cmdline, capsys):
    pattern_str = """
        [variables]
        thevar = "ANOTHERVAR"

        [scopes.ls]
        agent = "environment_variables"
        unset = ["SOMEVAR", "{{ variables.thevar }}"]
        export.LS_COLORS = "ace ventura"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    lines = out.splitlines()
    assert lines[0] == "unset SOMEVAR"
    assert lines[1] == "unset ANOTHERVAR"
    assert lines[2] == 'export LS_COLORS="ace ventura"'


def test_unset_string(dye_cmdline, capsys):
    pattern_str = """
        [scopes.unset]
        agent = "environment_variables"
        unset = "NOLISTVAR"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    lines = out.splitlines()
    assert lines[0] == "unset NOLISTVAR"



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable


from dye import Dye

SAMPLE_THEME = """
[colors]
foreground = "#f8f8f2"
foreground_high = "foreground"
foreground_medium = "foreground"
foreground_low = "foreground"

triad.first = "#aaaaaa"
triad.second = "#bbbbbb"
triad.third = "#cccccc"

[styles]
themeonly = "{{color.foreground}} on #393b47"
foreground = "{{ color.foreground }}"
text = "#d2d2d2 on #2d2d2d"
"""

SAMPLE_PATTERN = """
description = "Oxygen is a pattern with lots of space"
type = "dark"
version = "2.0"

requires_theme = "reqtheme"
prevent_themes = true

[colors]
pattern_purple =  "#bd93f8"
pattern_yellow =  "#f1fa8b"
notyet =  "{{ colors.yellow }}"

foreground =  "#e9e9e3"
background =  "#282a36"

foreground_unknown = "{{ colors.unknown }}"

background_high = "{{ colors.background }}"
background_medium = "{{ color.background }}"
background_low = "background"
background_double1 = "{{ color.background_low }}"
background_double2 = "background_medium"

yellow = "#e2eb9c"

triad.second = "#dd2222"

tetrad.first = "#ff1111"
tetrad.second = "#ff2222"
tetrad.third = "foreground_low"
tetrad.fourth = "{{ color.pattern_purple }}"
tetrad.fifth = "{{ colors.triad.third }}"

[styles]
notyet = "{{ styles.text }}"
foreground = "{{ color.foreground }}"
text = "bold {{ colors.foreground }} on {{ colors.background }}"
text_high = "{{ styles.text }}"
text_medium = "{{ style.text }}"
text_low = "text"
text_double1 = "{{ style.text_low }}"
text_double2 = "text_medium"
text_colorref = "{{ color.foreground_medium }}"
color1 = "#ff79c6"
color2 = "{{ colors.unknown }}"
color3 = "{{ style.unknown }}"
color4 = ""

pattern_text = '#cccccc on #ffffff'
pattern_text_high = '#000000 on #ffffff'
pattern_text_low = '#999999 on #ffffff'
pattern_yellow = "{{ colors.pattern_yellow }}"

triad_sty.first = "{{ colors.triad.first }}"
triad_sty.second = "pattern_text"
triad_sty.third = "{{ style.themeonly }}"
"""


#
# test the preview command
#
def test_preview_nothing(dye_cmdline, capsys):
    # no theme, no pattern is an error
    exit_code = dye_cmdline("preview")
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_ERROR
    assert not out
    assert "nothing to preview" in err


def test_preview(dye_cmdline, capsys):
    exit_code = dye_cmdline("preview", SAMPLE_THEME, SAMPLE_PATTERN)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert out
    assert not err
    # here's a list of strings that should be in the output
    tests = [
        "foreground",
        "pattern_purple",
        "themeonly",
        "yellow",
        "triad.second",
        "tetrad.fifth",
        "triad_sty.third",
    ]
    for test in tests:
        assert test in out


def test_theme_only(dye_cmdline, capsys):
    exit_code = dye_cmdline("preview", SAMPLE_THEME)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    tests = ["foreground_high", "triad.third", "text", "themeonly", "No pattern file."]
    for test in tests:
        assert test in out
    assert not err


def test_pattern_only(dye_cmdline, capsys):
    exit_code = dye_cmdline("preview", None, SAMPLE_PATTERN)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    tests = [
        "pattern_purple",
        "triad.second",
        "text",
        "triad_sty.third",
        "No theme file.",
    ]
    for test in tests:
        assert test in out
    assert not err


def test_no_text(dye_cmdline, capsys):
    theme_toml = """
        [colors]
        background = "#282a36"
        current_line = "#44475a"

        [styles]
        current_line =  "#f8f8f2 on #44475a"
        comment =  "{{color.comment}}"
    """
    exit_code = dye_cmdline("preview", theme_toml)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_ERROR
    assert not out
    assert "no 'text' style defined" in err


def test_empty_style(dye_cmdline, capsys):
    theme_toml = """
        [styles]
        text = "#d2d2d2 on #2d2d2d"
        empty_style = ""
    """
    exit_code = dye_cmdline("preview", theme_toml)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert "empty_style" in out
    assert not err



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""base class and mixin for all agents"""

import abc
import re

import rich.color

from ..exceptions import DyeError


class AgentBase(abc.ABC):
    """Abstract Base Class for all agents

    Subclass and implement `run()`. The first line of the class docstring
    is displayed by `shell-themer agents` as the description of the agent

    Creates a registry of all subclasses in cls.agents

    The string to use in your theme configuration file is the underscored
    class name, ie:

    EnvironmentVariables -> environment_variables
    LsColors -> ls_colors
    """

    classmap = {}

    @classmethod
    def __init_subclass__(cls):
        super().__init_subclass__()
        # make a registry of subclasses as they are defined
        cls.classmap[cls._name_of(cls.__name__)] = cls

    def __init__(self, scope):
        super().__init__()
        self.agent_name = self._name_of(self.__class__.__name__)
        self.scope = scope

    @classmethod
    def _name_of(cls, name: str) -> str:
        """Make an underscored, lowercase form of the given class name."""
        name = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", name)
        name = re.sub(r"([a-z\d])([A-Z])", r"\1_\2", name)
        name = name.replace("-", "_")
        return name.lower()

    @abc.abstractmethod
    def run(self, comments=False) -> str:
        """Do agent work. Anything returned will be sourced by the shell"""


class LsColorsFromStyle:
    """Generator mixin to create ls_colors type styles"""

    def ls_colors_from_style(self, name, style, mapp, scope_name, allow_unknown=False):
        """create an entry suitable for LS_COLORS from a style

        name should be a valid LS_COLORS entry, could be a code representing
        a file type, or a glob representing a file extension

        style is a style object

        mapp is a dictionary of friendly color names to native color names
            ie map['directory'] = 'di'

        allow_unknown - whether to allow [name] that is not in [mapp]. If false,
        an error will be generated if you use a [name] that is not in [mapp]. If
        true, [mapp] will only be used as a "friendly" name lookup.

        **msgdata - used for generating useful error messages
        prog = the name of the program
        scope_name = is the scope where this mapped occured

        returns a tuple of the mapped name and a phrase to add to LS_COLORS
        """
        ansicodes = ""
        if not style:
            return "", ""
        try:
            mapname = mapp[name]
        except KeyError as exc:
            if allow_unknown:
                # no problem if we didn't find name in mapp, we'll just use name
                # as is
                mapname = name
            else:
                # they used a style for a file attribute that isn't in the map
                # which is not allowed
                raise DyeError(
                    f"unknown style '{name}' while processing scope '{scope_name}'"
                ) from exc

        if style.color.type == rich.color.ColorType.DEFAULT:
            ansicodes = "0"
        else:
            # this works, but it uses a protected method
            #   ansicodes = style._make_ansi_codes(rich.color.ColorSystem.TRUECOLOR)
            # here's another approach, we ask the style to render a string, then
            # go peel the ansi codes out of the generated escape sequence
            ansistring = style.render("-----")
            # style.render uses this string to build it's output
            # f"\x1b[{attrs}m{text}\x1b[0m"
            # so let's go split it apart
            match = re.match(r"^\x1b\[([;\d]*)m", ansistring)
            # and get the numeric codes
            ansicodes = match.group(1)
        return mapname, f"{mapname}={ansicodes}"



#
"""Development related tasks to be run with 'invoke'"""

import contextlib
import os
import shutil

import invoke

#
# in the name namespace, the words are verbs, ie 'invoke test'
# in the check namespace, check is the verb, so the tasks are nouns 'invoke check tests'


# shared function
def rmrf(items, verbose=True):
    "Silently remove a list of directories or files"
    if isinstance(items, str):
        items = [items]

    for item in items:
        if verbose:
            print(f"Removing {item}")
        shutil.rmtree(item, ignore_errors=True)
        # rmtree doesn't remove bare files
        with contextlib.suppress(FileNotFoundError):
            os.remove(item)


# create namespaces
namespace = invoke.Collection()
namespace_clean = invoke.Collection("clean")
namespace.add_collection(namespace_clean, "clean")

namespace_check = invoke.Collection("check")
namespace.add_collection(namespace_check, "check")


#####
#
# testing, coverage, and quality
#
#####
@invoke.task
def test(context):
    "Run tests and code coverage using pytest"
    context.run("pytest", echo=True, pty=True)


namespace.add_task(test)
namespace_check.add_task(test, name="tests")


@invoke.task
def pytest_clean(context):
    "Remove pytest cache and code coverage files and directories"
    # pylint: disable=unused-argument
    dirs = [".pytest_cache", ".cache", ".coverage"]
    rmrf(dirs)


namespace_clean.add_task(pytest_clean, name="tests")


@invoke.task
def quality(context):
    "Check code quality using ruff"
    context.run("ruff check *.py src/dye tests", echo=True)


namespace.add_task(quality, name="inspect")
namespace_check.add_task(quality)


@invoke.task
def format_check(context):
    """Check if code is properly formatted using ruff"""
    context.run("ruff format --check *.py tests src", echo=True)


namespace_check.add_task(format_check, name="format")


@invoke.task
def formatt(context):
    """Format code using ruff"""
    context.run("ruff format *.py tests src", echo=True)


namespace.add_task(formatt, name="format")


#####
#
# build and publish
#
#####
DISTDIR = "dist"


@invoke.task
def dist_clean(context):
    "Remove the dist directory"
    # pylint: disable=unused-argument
    rmrf(DISTDIR)


namespace_clean.add_task(dist_clean, "dist")


@invoke.task
def eggs_clean(context):
    "Remove egg directories"
    # pylint: disable=unused-argument
    dirs = set()
    dirs.add(".eggs")
    for name in os.listdir(os.curdir):
        if name.endswith(".egg-info"):
            dirs.add(name)
        if name.endswith(".egg"):
            dirs.add(name)
    rmrf(dirs)


namespace_clean.add_task(eggs_clean, "eggs")


@invoke.task
def bytecode_clean(context):
    "Remove __pycache__ directories and *.pyc files"
    # pylint: disable=unused-argument
    dirs = set()
    for root, dirnames, files in os.walk(os.curdir):
        if "__pycache__" in dirnames:
            dirs.add(os.path.join(root, "__pycache__"))
        for file in files:
            if file.endswith(".pyc"):
                dirs.add(os.path.join(root, file))
    print("Removing __pycache__ directories and .pyc files")
    rmrf(dirs, verbose=False)


namespace_clean.add_task(bytecode_clean, "bytecode")


@invoke.task(pre=list(namespace_check.tasks.values()), default=True)
def check_all(context):
    "Run this before you commit or submit a pull request"
    # pylint: disable=unused-argument


namespace_check.add_task(check_all, "all")


@invoke.task(pre=list(namespace_clean.tasks.values()), default=True)
def clean_all(context):
    "Clean everything"
    # pylint: disable=unused-argument


namespace_clean.add_task(clean_all, "all")


@invoke.task(pre=[clean_all])
def build(context):
    "Create a distribution"
    context.run("uv build")


namespace.add_task(build)


@invoke.task(pre=[build])
def pypi(context):
    "Build and publish a distribution to pypi"
    context.run("uv publish")


namespace.add_task(pypi)


@invoke.task(pre=[build])
def pypi_test(context):
    "Build and publish a distribution to https://test.pypi.org"
    context.run("uv publish --index test-pypi")


namespace.add_task(pypi_test)



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""LS_COLORS agent"""

import rich.style

from ..exceptions import DyeSyntaxError
from .base import AgentBase, LsColorsFromStyle


class LsColors(AgentBase, LsColorsFromStyle):
    "Create LS_COLORS environment variable for use with GNU ls"

    LS_COLORS_BASE_MAP = {
        # map both a friendly name and the "real" name
        "text": "no",
        "file": "fi",
        "directory": "di",
        "symlink": "ln",
        "multi_hard_link": "mh",
        "pipe": "pi",
        "socket": "so",
        "door": "do",
        "block_device": "bd",
        "character_device": "cd",
        "broken_symlink": "or",
        "missing_symlink_target": "mi",
        "setuid": "su",
        "setgid": "sg",
        "sticky": "st",
        "other_writable": "ow",
        "sticky_other_writable": "tw",
        "executable_file": "ex",
        "file_with_capability": "ca",
    }
    # this map allows you to either use the 'native' color code, or the
    # 'friendly' name defined by shell-themer
    LS_COLORS_MAP = {}
    for friendly, actual in LS_COLORS_BASE_MAP.items():
        LS_COLORS_MAP[friendly] = actual
        LS_COLORS_MAP[actual] = actual

    def run(self, comments=False):
        "Render a LS_COLORS variable suitable for GNU ls"
        outlist = []
        havecodes = []
        # figure out if we are clearing builtin styles
        try:
            clear_builtin = self.scope.definition["clear_builtin"]
            if not isinstance(clear_builtin, bool):
                raise DyeSyntaxError(
                    f"scope '{self.scope}' requires 'clear_builtin' to be true or false"
                )
        except KeyError:
            clear_builtin = False

        # iterate over the styles given in our configuration
        for name, style in self.scope.styles.items():
            if style:
                mapcode, render = self.ls_colors_from_style(
                    name,
                    style,
                    self.LS_COLORS_MAP,
                    self.scope.name,
                    allow_unknown=False,
                )
                havecodes.append(mapcode)
                outlist.append(render)

        if clear_builtin:
            style = rich.style.Style.parse("default")
            # go through all the color codes, and render them with the
            # 'default' style and add them to the output
            for name, code in self.LS_COLORS_BASE_MAP.items():
                if code not in havecodes:
                    _, render = self.ls_colors_from_style(
                        name,
                        style,
                        self.LS_COLORS_MAP,
                        allow_unknown=False,
                        scope_name=self.scope,
                    )
                    outlist.append(render)

        # process the filesets

        # figure out which environment variable to put it in
        try:
            varname = self.scope.definition["environment_variable"]
        except KeyError:
            varname = "LS_COLORS"

        # even if outlist is empty, we have to set the variable, because
        # when we are switching a theme, there may be contents in the
        # environment variable already, and we need to tromp over them
        # we chose to set the variable to empty instead of unsetting it
        return f'''export {varname}="{":".join(outlist)}"'''



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

from unittest import mock

from dye import Dye


#
# test the themes command
#
def test_themes(dye_cmdline, capsys, mocker, tmp_path):
    # gotta patch dye_dir, which reads the environment variable DYE_DIR
    dirmock = mocker.patch(
        "dye.Dye.dye_dir", create=True, new_callable=mock.PropertyMock
    )
    dirmock.return_value = tmp_path
    # make sure the themes subdirectory exists
    theme_dir = tmp_path / "themes"
    theme_dir.mkdir()

    # write some empty toml files into the directory
    bases = ["one", "two", "three"]
    for base in bases:
        path = theme_dir / f"{base}.toml"
        with open(path, "w", encoding="utf8") as fvar:
            fvar.write(f"# a toml file for theme {base}")
    # now go run the command, which should list the themes
    exit_code = dye_cmdline("themes")
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    for base in bases:
        assert base in out


def test_themes_no_theme_dir(dye_cmdline, capsys, mocker, tmp_path):
    # gotta patch dye_dir, which reads the environment variable DYE_DIR
    dirmock = mocker.patch(
        "dye.Dye.dye_dir", create=True, new_callable=mock.PropertyMock
    )
    dirmock.return_value = tmp_path
    # now go run the command, which should list the themes
    exit_code = dye_cmdline("themes")
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_ERROR
    assert not out
    assert "is not a directory" in err


def test_themes_no_dye_dir(dye_cmdline, capsys, mocker):
    # gotta patch dye_dir, which reads the environment variable DYE_DIR
    dirmock = mocker.patch(
        "dye.Dye.dye_dir", create=True, new_callable=mock.PropertyMock
    )
    dirmock.return_value = None
    # now go run the command, which should list the themes
    exit_code = dye_cmdline("themes")
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_ERROR
    assert not out
    assert "DYE_DIR environment variable must be set" in err



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""environment variables agent"""

from .base import AgentBase


class EnvironmentVariables(AgentBase):
    "Export and unset environment variables"

    def run(self, comments=False) -> str:
        output = []
        try:
            unsets = self.scope.definition["unset"]
            if isinstance(unsets, str):
                # if they used a string in the config file instead of a list
                # process it like a single item instead of trying to process
                # each letter in the string
                unsets = [unsets]
            for var in unsets:
                output.append(f"unset {var}")
        except KeyError:
            # no unsets
            pass
        # render the variables to export
        try:
            exports = self.scope.definition["export"]
            for var, value in exports.items():
                output.append(f'export {var}="{value}"')
        except KeyError:
            # no exports
            pass
        return "\n".join(output)
